const { STATUS_CODE, TEMP_IMAGE_DIR, TEMP_VIDEO_DIR, TEMP_AUDIO_DIR } = require("../helper/constants")
const { generateCanvas, mergeImages } = require("../helper/tutor-shot")
const path = require('path')
const sharp = require('sharp')
const fs = require('fs')
const { createVideoFromImages,createVideo, joinAudioVideo } = require("../helper/ffmpeg-helper")
const { deleteFile, downloadFile } = require("../helper/utilities")
const { generatePreSignedGetUrl,uploadThumbnailToAwsS3, uploadToAwsS3 } = require("../helper/upload")
const mongoose = require('mongoose')
const ObjectId = mongoose.Types.ObjectId
const ffmpeg = require("fluent-ffmpeg");
const CONFIG = require("../config.js")
const { execSync } = require('child_process');
const socket = require("../socket.js")

module.exports.generateVideo = async function (req, res) {
    let uid;
    let _id
    let filesCreated = []
    let socketPayload
    try {
        let tutorShotData = req.body.tutorShotData;
        let videoPath; // local path of the video file
        let thumbnailPath // local path of the thumbnail file
        let fileName; // remote name of the final video file
        let thumbnailName; // remote name of the final thumbnail

        if (tutorShotData) {
            uid = tutorShotData.uid
            _id = tutorShotData._id
            socketPayload = {
                uid: tutorShotData.uid,
                _id: tutorShotData._id,
                task: "Video Generation",
                percent: 0,
                subTask: [{ task: "Image Compiled", percent: 0 }]
            }
            if (tutorShotData?.audio?.length) {
                socketPayload.subTask.push({ task: "Audio mixing", percent: 0 })
            }

            let imageDir = TEMP_IMAGE_DIR
            let imagePathString = path.join(imageDir, `${tutorShotData._id}_%d.png`)

            videoPath = path.join(TEMP_VIDEO_DIR, `${tutorShotData._id}.webm`)
            thumbnailPath = path.join(TEMP_IMAGE_DIR, `${tutorShotData._id}_thumbnail`)
            fileName = `${tutorShotData._id}.webm`
            thumbnailName = `${tutorShotData._id}_thumbnail.png`

            // // this is to generate each individual images and get their buffers
            let frameNumber = 1;
            let imageArr = []

            socket.emit("videoGenerationProgress", socketPayload)

            for (let i = 0; i < tutorShotData.segments.length; i++) {
                if (tutorShotData.segments[i].imageUrl) tutorShotData.segments[i].signedImageUrl = await generatePreSignedGetUrl(tutorShotData.segments[i].imageUrl)
                let CE = generateCanvas(tutorShotData.segments[i], { dimension: tutorShotData.segments[i].dimension })
                let mergedImage = await mergeImages(
                    [
                        { url: tutorShotData.segments[i].signedImageUrl },
                        { buffer: await CE.getBuffer() }
                    ],
                    {
                        outputType: "buffer",
                        dimension: tutorShotData.segments[i].dimension
                    }
                )
                let limit = ((tutorShotData.segments[i]?.duration) ? (tutorShotData.segments[i].duration) : 2);
                let frameNum = 0
                let firstImage = path.join(imageDir, `${tutorShotData._id}_${frameNumber++}.png`)
                await sharp(mergedImage).toFile(firstImage)
                imageArr.push(firstImage)
                filesCreated.push(firstImage)
                // do {
                //     let fileName = path.join(imageDir, `${tutorShotData._id}_${frameNumber++}.png`)
                //     imageArr.push(fileName)
                //     filesCreated.push(fileName)
                //     fs.copyFileSync(firstImage, fileName)
                // } while (++frameNum < limit)

                if (i == 0 && thumbnailPath) {
                    fs.copyFileSync(firstImage, thumbnailPath)
                    filesCreated.push(thumbnailPath)
                }

                socketPayload.subTask[0].percent = Math.round(((i + 1) / tutorShotData.segments.length) * 75); // 75% for generating images
                socketPayload.percent = Math.round(tutorShotData?.audio?.length ? socketPayload.subTask[0].percent / 2 : socketPayload.subTask[0].percent);
                socket.emit("videoGenerationProgress", socketPayload);
            }


            let options = {
                outputFilePath: path.dirname(videoPath),
                outputFileName: path.parse(videoPath).name,
                outputFormat: path.extname(videoPath).slice(1),
                frameRate: 24,
                imagePathString: imagePathString,
                deleteImages: true,// default is true
            }
            await createVideo(imageArr,videoPath)
            // await createVideoFromImages([], options);
            filesCreated.push(path.dirname(videoPath))
            socketPayload.subTask[0].percent = 100; // 100% if video(without audio) is generated
            socketPayload.percent = Math.round(tutorShotData?.audio?.length ? socketPayload.subTask[0].percent / 2 : socketPayload.subTask[0].percent);
            socket.emit("videoGenerationProgress", socketPayload);

            !(async function () {
                deleteFile(imageArr)
            })()

            // // to save audio
            if (tutorShotData.audio?.length) {
                socketPayload.percent = 50
                socketPayload.subTask[0].percent = 50
                socket.emit("videoGenerationProgress", socketPayload)
                let audioPath = path.join(TEMP_AUDIO_DIR, `ts-${tutorShotData._id}.mp3`); // path to only audio
                await createAudio(tutorShotData.audio, audioPath)
                finalVideoPath = path.join(TEMP_VIDEO_DIR, `ts-final-${tutorShotData._id}.webm`); // path to combined video and audio
                 await joinAudioVideo(videoPath, audioPath, finalVideoPath)
                 videoPath = finalVideoPath;
            } else {
                socketPayload.percent = 99
                socketPayload.subTask[0].percent = 99
                socket.emit("videoGenerationProgress", socketPayload)
            }

            // uploading process
            let remotePath = path.join(CONFIG.S3_TUTOR_SHOT, fileName)
            let remoteThumbnailPath = path.join(CONFIG.S3_TUTOR_SHOT, thumbnailName)

            let url = await uploadToAwsS3(videoPath, remotePath)
            await uploadToAwsS3(thumbnailPath, remoteThumbnailPath)
            tutorShotData.thumbnailUrl = remoteThumbnailPath
            tutorShotData.signedVideoUrl = await generatePreSignedGetUrl(remotePath)
            tutorShotData.signedThumbnailUrl = await generatePreSignedGetUrl(remoteThumbnailPath)
            tutorShotData.videoUrl = remotePath
            tutorShotData.generatingVideo = false

            socketPayload.tutorShotData = {
                videoUrl: remotePath,
                thumbnailUrl: remoteThumbnailPath,
                signedVideoUrl: tutorShotData?.signedVideoUrl,
                signedThumbnailUrl: tutorShotData?.signedThumbnailUrl,
            }
            socketPayload.percent = 100
            socketPayload.complete = true
            socket.emit("videoGenerationProgress", socketPayload)
            console.log('---Tutor-shot-generated----', tutorShotData?._id)
        }

        res.status(STATUS_CODE.OK).json(tutorShotData)
    } catch (err) {
        console.log('------ERROR:GENERATING VIDEO', err)
        console.log('------ERROR:GENERATING VIDEO')
        res.status(STATUS_CODE.ERROR).json()
        socket.emit("videoGenerationFailed", { uid, _id })
        deleteFile(filesCreated)
    }
}
async function createAudio(audioList, audioPath) {
    let downloadAudioPromise = []
    let tempFiles = []
    for (let a = 0; a < audioList.length; a++) {
        if (!audioList[a]?._id) {
            audioList[a]._id = (new ObjectId()).toString()
        }
        audioList[a].localPath = path.join(TEMP_AUDIO_DIR, `${audioList[a]._id}.mp3`)
        audioList[a].trimmedLocalPath = path.join(TEMP_AUDIO_DIR, `${audioList[a]._id}_trimmed.mp3`)
        downloadAudioPromise.push(new Promise(async (response, reject) => {
            try {
                audioList[a].url = await generatePreSignedGetUrl(audioList[a].remotePath, audioList[a].localPath)
                await downloadFile(audioList[a].url, audioList[a].localPath)
                tempFiles.push(audioList[a].localPath)
                if (audioList[a].trim) {
                    ffmpeg()
                        .input(audioList[a].localPath)
                        .seekInput(0)
                        .duration(audioList[a].trim)
                        .audioCodec('copy') // Copy audio codec without re-encoding
                        .output(audioList[a].trimmedLocalPath)
                        .on('end', () => {
                            console.log('Trimming finished!');
                            response(audioList[a].localPath)
                        })
                        .on('error', (err) => {
                            console.error('Error:', err);
                            reject(err)
                        })
                        .run();

                } else {
                    response(audioList[a].localPath)
                }
            } catch (err) {
                reject(err)
            }
        }))
    }

    await Promise.all(downloadAudioPromise)


    await new Promise((res, rej) => {

        let audioObj = ffmpeg()
        audioObj.on('end', () => {
            deleteFile(tempFiles)
            tempFiles.push(audioPath)
            res()
        }).on('error', (err) => {
            deleteFile(tempFiles)
            console.error('----------Error:\n', err);
            rej(err)
        })

        for (let a = 0; a < audioList.length; a++) {
            let prevEndTime = 0
            if (a != 0) prevEndTime = audioList[a - 1].trim ? audioList[a - 1].trim : (audioList[a - 1].endTime)
            if (prevEndTime) {
                const silenceAudio = path.join(TEMP_AUDIO_DIR, `temp-${audioList[a]._id}.mp3`);
                let waitTime = audioList[a].startTime - prevEndTime; // in seconds
                if (audioList[a].trim) {
                    waitTime = audioList[a].trim
                }
                let command = `ffmpeg -f lavfi -i anullsrc=r=44100 -t ${waitTime} -c:a libmp3lame -q:a 0 ${silenceAudio} -y`
                execSync(command)
                audioObj.input(silenceAudio).inputOptions('-ss', '0');
                tempFiles.push(silenceAudio)
            }
            audioObj.input(audioList[a].trim ? audioList[a].trimmedLocalPath : audioList[a].localPath).inputOptions('-ss', '0');
        }

        audioObj.mergeToFile(audioPath, path.dirname(audioPath))
    })

    return
}