const fs = require('fs')
const axios = require("axios")

module.exports.createDirectory = function (directory) {
    fs.mkdir(directory, { recursive: true }, (err) => {
        if (err) {
            console.error(`Error creating directory: ${err}`);
        } else {
            console.log(`Directory created successfully at: ${directory}`);
        }
    });
}


module.exports.deleteFile = function (item) {

    function deleteFileByPath(path) {
        fs.unlink(path, (err) => { if (err) console.log(err) })
    }

    if (typeof item == 'string') {
        deleteFileByPath(item)
    } else if (Array.isArray(item)) {
        for (let i = 0; i < item.length; i++) {
            if (typeof item[i] == 'string') deleteFileByPath(item[i])
        }
    }
    return
}


module.exports.downloadFile = async function (url, localPath) {
    try {
        const response = await axios({
            method: 'GET',
            url: url,
            responseType: 'stream', // Important to handle binary data
        });

        // Create a writable stream and pipe the response data to it
        const writer = fs.createWriteStream(localPath);
        response.data.pipe(writer);

        return new Promise((resolve, reject) => {
            writer.on('finish', resolve);
            writer.on('error', reject);
        });
    } catch (error) {
        console.error('Error:', error.message);
        throw error;
    }
}